import itertools
from operator import itemgetter


def sort(items: list, *keys):
    return sorted(items, key=itemgetter(*keys))


def group_by(items: list, *keys):

    results = []
    for key_values, items_iterator in itertools.groupby(sort(items, *keys), key=itemgetter(*keys)):
        result = dict([x for x in zip(keys, key_values if isinstance(key_values, tuple) else (key_values,))])
        result.update({'items': [x for x in items_iterator]})
        results.append(result)
    return results


def count(items: list, *keys):

    results = []
    for item in group_by(items, *keys):
        result = {key: item[key] for key in keys}
        result.update({'count': len(item['items'])})
        results.append(result)
    return results



def compare_instances(x, y, instance_type) -> bool:
    return isinstance(x, instance_type) and isinstance(y, instance_type)


def list_diff(x: list, y: list, assert_only: bool = False) -> list:

    errors: list = []

    for idx, xv in enumerate(x):

        error_obj: dict = dict(x=xv)

        try:
            yv = y[idx]
        except IndexError:
            error_obj['type'] = 'MISSING'
            error_obj['field'] = '[{}]'.format(idx)
            error_obj['message'] = '[{}]: not found'.format(idx)
            errors.append(error_obj)
            continue

        if isinstance(xv, dict) and isinstance(yv, dict):
            sub_errors = json_diff(xv, yv)
            for error in sub_errors:
                error['field'] = '[{}].{}'.format(idx, error['field'])
                error['message'] = '[{}].{}'.format(idx, error['message'])
                errors.append(error)
            continue

        if xv != yv:
            error_obj['type'] = 'MISMATCH'
            error_obj['y'] = yv
            format_strings = dict(idx=idx, x=xv, type_x=type(xv).__name__, y=yv, type_y=type(yv).__name__)
            error_obj['message'] = '[{idx}]: {x} ({type_x}) != {y} ({type_y})'.format(**format_strings)
            errors.append(error_obj)

    if assert_only:
        return errors

    for idx, yv in enumerate(y):

        error_obj = {}

        if idx >= len(x):
            error_obj['y'] = yv
            error_obj['type'] = 'MISSING'
            error_obj['field'] = '[{}]'.format(idx)
            error_obj['message'] = '[{}]: not found'.format(idx)
            errors.append(error_obj)

    return errors


def compare_field(obj, field: str, value, ignore_missing_y_fields: bool) -> list:

    error = dict(field=field, x=value)

    if field not in obj:
        error['type'] = 'MISSING'
        error['message'] = '{}: not found'.format(field)
        return [error]

    if compare_instances(value, obj[field], dict):
        errors = json_diff(value, obj[field], ignore_missing_y_fields)
        for item in errors:
            item['field'] = '{}.{}'.format(field, item['field'])
            item['message'] = '{}.{}'.format(field, item['message'])
        return errors

    if compare_instances(value, obj[field], list):
        errors = list_diff(value, obj[field], ignore_missing_y_fields)
        for item in errors:
            item['field'] = '{}{}'.format(field, item['field']) if 'field' in item else field
            item['message'] = '{}{}'.format(field, item['message'])
        return errors

    if value != obj[field]:
        error['type'] = 'MISMATCH'
        error['y'] = obj[field]
        format_strings = dict(key=field, x=value, type_x=type(value).__name__,
                              y=obj[field], type_y=type(obj[field]).__name__)
        error['message'] = '{key}: {x} ({type_x}) != {y} ({type_y})'.format(**format_strings)
        return [error]

    return []


def json_diff(x: dict, y: dict, assert_only: bool = False) -> list:

    errors: list = []

    for k in x:
        errors += compare_field(y, k, x[k], assert_only)

    if assert_only:
        return errors

    for k in y:
        if k not in x:
            errors.append(dict(field=k, y=y[k], type='MISSING',
                               message='{}: not found'.format(k)))

    return errors


def compare(x, y, assert_only=False):

    if isinstance(x, dict) and isinstance(y, dict):
        return json_diff(x, y, assert_only=assert_only)

    if isinstance(x, list) and isinstance(y, list):
        return list_diff(x, y, assert_only=assert_only)



import json

import click
from rich.table import Table
from rich.console import Console

from melano.diff import compare


def read_file(path):

    with open(path, 'r') as fp:
        return json.load(fp)


def format_value(value):

    if not value:
        return 'null'

    if isinstance(value, (int, float)):
        return str(value)

    if isinstance(value, bool):
        return 'true' if value else 'false'

    if isinstance(value, dict):
        return json.dumps(value)

    return value


@click.group()
def melano():
    pass


@melano.command()
@click.argument('file1')
@click.argument('file2')
@click.option('--assert-only/--no-assert-only', default=False)
def diff(file1, file2, assert_only):

    diffs = compare(read_file(file1), read_file(file2), assert_only)

    console = Console()
    table = Table(title="Differences", show_lines=True)

    table.add_column('s.no.', justify="center")
    # table.add_column("type", justify="center")
    table.add_column("field")
    table.add_column(file1)
    table.add_column(file2)
    # table.add_column("Message")

    for idx, diff in enumerate(diffs):
        x = format_value(diff.get('x', '-'))
        y = format_value(diff.get('y', '-'))
        color = 'red' if diff['type'] == 'MISSING' else 'yellow'
        table.add_row(
            str(idx+1),
            # diff['type'],
            diff['field'],
            x,
            y,
            # diff['message'],
            style=color)

    console.print(table)


if __name__ == '__main__':
    melano()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['melano']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0', 'rich>=10.1.0,<11.0.0']

entry_points = \
{'console_scripts': ['melano = melano.cli:melano']}

setup_kwargs = {
    'name': 'melano',
    'version': '0.1.0',
    'description': 'JSON utility',
    'long_description': "# melano\n\n`melano` is JSON utility library\n\n## features\n### dict compare\n\n```shell\n>>> from melano import diff\n>>> x, y = dict(x=1, y=2), dict(x=2, z=1)\n>>> diff.compare(x, y)\n[{'field': 'x', 'x': 1, 'type': 'MISMATCH', 'y': 2, 'message': 'x: 1 (int) != 2 (int)'}, {'field': 'y', 'x': 2, 'type': 'MISSING', 'message': 'y: not found'}, {'field': 'z', 'y': 1, 'type': 'MISSING', 'message': 'z: not found'}]\n```\n\n### list of dict operations\n#### sort \n\n```shell\n>>> from melano import sort\n>>> items = [dict(x=1, y=5), dict(x=4, y=2), dict(x=2, y=3), dict(x=1, y=2)]\n>>> sort(items, 'x', 'y')\n[{'x': 1, 'y': 2}, {'x': 1, 'y': 5}, {'x': 2, 'y': 3}, {'x': 4, 'y': 2}]\n```\n\n#### group by\n\n```shell\n>>> from melano import group_by\n>>> items = [dict(x=1, y=5, z=5), dict(x=4, y=2, z=3), dict(x=4, y=2, z=10)]\n>>> group_by(items, 'x', 'y')\n[{'x': 1, 'y': 5, 'items': [{'x': 1, 'y': 5, 'z': 5}]}, {'x': 4, 'y': 2, 'items': [{'x': 4, 'y': 2, 'z': 3}, {'x': 4, 'y': 2, 'z': 10}]}]\n```\n\n#### count\n\n```shell\n>>> from melano import count\n>>> items = [dict(x=1, y=5, z=5), dict(x=4, y=2, z=3), dict(x=4, y=2, z=10)]\n>>> count(items, 'x', 'y')\n[{'x': 1, 'y': 5, 'count': 1}, {'x': 4, 'y': 2, 'count': 2}]\n```",
    'author': 'Suganthan Sundararaju',
    'author_email': 'suganthsundar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/suganthsundar/melano',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
